module APL.Interp_Tests (tests) where

import APL.AST (Exp (..))
import APL.Eval (eval)
import APL.InterpIO (runEvalIO)
import APL.InterpPure (runEval)
import APL.Monad
import APL.Util (captureIO)
import Test.Tasty (TestTree, testGroup)
import Test.Tasty.HUnit (testCase, (@?=))

eval' :: Exp -> ([String], Either Error Val)
eval' = runEval . eval

evalIO' :: Exp -> IO (Either Error Val)
evalIO' = runEvalIO . eval

tests :: TestTree
tests = testGroup "Free monad interpreters" [pureTests, ioTests]

pureTests :: TestTree
pureTests =
  testGroup
    "Pure interpreter"
    [ testCase "localEnv" $
        runEval
          ( localEnv (const [("x", ValInt 1)]) $
              askEnv
          )
          @?= ([], Right [("x", ValInt 1)]),
      --
      testCase "Let" $
        eval' (Let "x" (Add (CstInt 2) (CstInt 3)) (Var "x"))
          @?= ([], Right (ValInt 5)),
      --
      testCase "Let (shadowing)" $
        eval'
          ( Let
              "x"
              (Add (CstInt 2) (CstInt 3))
              (Let "x" (CstBool True) (Var "x"))
          )
          @?= ([], Right (ValBool True)),
      --
      testCase "State" $
        runEval
          ( do
              putState [(ValInt 0, ValInt 1)]
              modifyState $ map (\(key, _) -> (key, ValInt 5))
              getState
          )
          @?= ([], Right [(ValInt 0, ValInt 5)]),
      --
      testCase "Print" $
        runEval (evalPrint "test")
          @?= (["test"], Right ()),
      --
      testCase "Error" $
        runEval
          ( do
              _ <- failure "Oh no!"
              evalPrint "test"
          )
          @?= ([], Left "Oh no!"),
      --
      testCase "Div0" $
        eval' (Div (CstInt 7) (CstInt 0))
          @?= ([], Left "Division by zero"),
      --
      testCase "TryCatch try1" $
        eval' (TryCatch (CstInt 1) (CstInt 2))
          @?= ([], Right (ValInt 1)),
      --
      testCase "TryCatch try2" $
        eval' (TryCatch (CstInt 1) (Div (CstInt 1) (CstInt 0)))
          @?= ([], Right (ValInt 1)),
      --
      testCase "TryCatch catch1" $
        eval' (TryCatch (Div (CstInt 1) (CstInt 0)) (CstInt 1))
          @?= ([], Right (ValInt 1)),
      --
      testCase "TryCatch catch2" $
        eval' (TryCatch (Div (CstInt 1) (CstInt 0)) (Div (CstInt 1) (CstInt 0)))
          @?= ([], Left "Division by zero"),
      --
      testCase "KvPutOp" $
        eval' (KvPut (CstInt 1) (CstInt 2))
          @?= ([], Right (ValInt 2)),
      --
      testCase "KvGetOp" $
        eval' (Let "_" (KvPut (CstInt 1) (CstInt 2)) (KvGet (CstInt 1)))
          @?= ([], Right (ValInt 2)),
      --
      testCase "KvGetOp shadowing" $
        eval' (Let "_" (KvPut (CstInt 1) (CstInt 2)) (Let "_" (KvPut (CstInt 1) (CstInt 3)) (KvGet (CstInt 1))))
          @?= ([], Right (ValInt 3)),
      --
      testCase "KvGetOp fail" $
        eval' (KvGet (CstInt 1))
          @?= ([], Left "Cannot find key: ValInt 1"),
      --
      testCase "TransactionOp 1" $
        let goodPut = (evalKvPut (ValInt 0) (ValInt 1)) in
        runEval (transaction goodPut)
          @?= ([], Right ()),
      --
      testCase "TransactionOp 2" $
        let goodPut = evalKvPut (ValInt 0) (ValInt 1) in
        let get0 = eval (KvGet (CstInt 0)) in
        runEval (transaction goodPut >> get0)
          @?= ([], Right (ValInt 1)),
      --
      testCase "TransactionOp 3" $
        let oPut = (evalKvPut (ValInt 1) (ValInt 2)) in
        let okayPut = ((evalKvGet (ValInt 1) >> evalKvPut (ValInt 0) (ValInt 1))) in
        let get0 = eval (KvGet (CstInt 0)) in
        runEval (oPut >> (transaction okayPut) >> get0)
          @?= ([], Right (ValInt 1)),
      --
      testCase "TransactionOp Fail" $
        let badPut = (evalKvPut (ValInt 0) (ValInt 1)) >> (failure "die") in
        let get0 = eval (KvGet (CstInt 0)) in
        runEval (transaction badPut >> get0)
          @?= ([], Left "Cannot find key: ValInt 0"),
      --
      testCase "TransactionOp Propagation" $
        let badPut = (evalKvPut (ValInt 0) (ValInt 1)) >> (failure "die") in
        runEval (transaction badPut)
          @?= ([], Right ()),
      --
      testCase "TransactionOp Printing" $
        runEval (transaction (evalPrint "weee" >> failure "oh shit"))
          @?= (["weee"], Right ()),
      --
      testCase "TransactionOp Nested" $
        let badPut = (evalKvPut (ValInt 0) (ValInt 1)) >> (failure "die") in
        let goodPut = evalKvPut (ValInt 0) (ValInt 1) in
        let get0 = eval (KvGet (CstInt 0)) in
        runEval (transaction (goodPut >> transaction badPut) >> get0)
          @?= ([], Right (ValInt 1))
    ]

ioTests :: TestTree
ioTests =
  testGroup
    "IO interpreter"
    [ testCase "print" $ do
        let s1 = "Lalalalala"
            s2 = "Weeeeeeeee"
        (out, res) <-
          captureIO [] $
            runEvalIO $ do
              evalPrint s1
              evalPrint s2
        (out, res) @?= ([s1, s2], Right ()),
      ---
      testCase "print 2" $ do
          (out, res) <-
            captureIO [] $
              evalIO' $
                Print "This is also 1" $
                  Print "This is 1" $
                    CstInt 1
          (out, res) @?= (["This is 1: 1", "This is also 1: 1"], Right $ ValInt 1),
      ---
      testCase "State" $ do
        r <- runEvalIO $ do
          putState [(ValInt 0, ValInt 1)]
          modifyState $ map (\(key, _) -> (key, ValInt 5))
          getState

        r @?= Right [(ValInt 0, ValInt 5)],
      ---
      testCase "TryCatch try1" $ do
        r <- evalIO' (TryCatch (CstInt 1) (CstInt 2))
        r @?= Right (ValInt 1),
      ---
      testCase "TryCatch try2" $ do
        r <- evalIO' (TryCatch (CstInt 1) (Div (CstInt 1) (CstInt 0)))
        r @?= Right (ValInt 1),
      ---
      testCase "TryCatch catch1" $ do
        r <- evalIO' (TryCatch (Div (CstInt 1) (CstInt 0)) (CstInt 1))
        r @?= Right (ValInt 1),
      ---
      testCase "TryCatch catch2" $ do
        r <- evalIO' (TryCatch (Div (CstInt 1) (CstInt 0)) (Div (CstInt 1) (CstInt 0)))
        r @?= Left "Division by zero",
      ---
      testCase "KvPutOp" $ do
        r <- evalIO' (KvPut (CstInt 1) (CstInt 2))
        r @?= Right (ValInt 2),
      --
      testCase "KvGetOp" $ do
        r <- evalIO' (Let "_" (KvPut (CstInt 1) (CstInt 2)) (KvGet (CstInt 1)))
        r @?= Right (ValInt 2),
      --
      testCase "KvGetOp Bool" $ do
        r <- evalIO' (Let "_" (KvPut (CstBool True) (CstBool False)) (KvGet (CstBool True)))
        r @?= Right (ValBool False),
      --
      testCase "KvGetOp shadowing" $ do
        r <- evalIO' (Let "_" (KvPut (CstInt 1) (CstInt 2)) (Let "_" (KvPut (CstInt 1) (CstInt 3)) (KvGet (CstInt 1))))
        r @?= Right (ValInt 3),
      --
      testCase "KvGetOp fail" $ do
        (_, r) <-
          captureIO [":)"] $
            evalIO' (KvGet (CstInt 1))
        r @?= Left "Invalid key: :)",
      --
      testCase "KvGetOp invalid int" $ do
        (_, r) <-
          captureIO ["ValInt 1"] $
            evalIO' (Let "_" (KvPut (CstInt 1) (CstInt 2)) (KvGet (CstInt 3)))
        r @?= Right (ValInt 2),
      --
      testCase "KvGetOp invalid bool" $ do
        (_, r) <-
          captureIO ["ValBool True"] $
            evalIO' (Let "_" (KvPut (CstBool True) (CstInt 2)) (KvGet (CstInt 3)))
        r @?= Right (ValInt 2),
      --
      testCase "KvGetOp invalid multiple" $ do
        (_, r) <-
          captureIO ["ValInt 1","ValInt 2","ValInt 3"] $
            evalIO' (Let "_" (KvPut (CstInt 3) (CstInt 2)) (KvGet (CstInt 4)))
        r @?= Right (ValInt 2),
      --
      testCase "TransactionOp 1" $
        let goodPut = (evalKvPut (ValInt 0) (ValInt 1)) in do
        r <- runEvalIO (transaction goodPut)
        r @?= Right (),
      --
      testCase "TransactionOp 2" $
        let goodPut = evalKvPut (ValInt 0) (ValInt 1) in
        let get0 = eval (KvGet (CstInt 0)) in do
        r <- runEvalIO (transaction goodPut >> get0)
        r @?= Right (ValInt 1),
      --
      testCase "TransactionOp 3" $
        let oPut = (evalKvPut (ValInt 1) (ValInt 2)) in
        let okayPut = ((evalKvGet (ValInt 1) >> evalKvPut (ValInt 0) (ValInt 1))) in
        let get0 = eval (KvGet (CstInt 0)) in do
        r <- runEvalIO (oPut >> (transaction okayPut) >> get0)
        r @?= Right (ValInt 1),
      --
      testCase "TransactionOp Fail" $
        let badPut = (evalKvPut (ValInt 0) (ValInt 1)) >> (failure "die") in
        let get0 = eval (KvGet (CstInt 0)) in do
        (_, r) <-
          captureIO [":)"] $
            runEvalIO (transaction badPut >> get0)
        r @?= Left "Invalid key: :)",
      -- --
      testCase "TransactionOp Propagation" $
        let badPut = (evalKvPut (ValInt 0) (ValInt 1)) >> (failure "die") in do
        r <- runEvalIO (transaction badPut)
        r @?= Right (),
      -- --
      testCase "TransactionOp Printing" $ do
        (p, r) <-
          captureIO [] $
            runEvalIO (transaction (evalPrint "weee" >> failure "oh shit"))
        (p, r) @?= (["weee"], Right ()),
      -- --
      testCase "TransactionOp Nested" $
        let badPut = (evalKvPut (ValInt 0) (ValInt 1)) >> (failure "die") in
        let goodPut = evalKvPut (ValInt 0) (ValInt 1) in
        let get0 = eval (KvGet (CstInt 0)) in do
        r <- runEvalIO (transaction (goodPut >> transaction badPut) >> get0)
        r @?= Right (ValInt 1)
    ]
